"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Image subclass:#Depth8Image
	 instanceVariableNames:''
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Graphics-Images'
!

Depth8Image comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

$Header: /cvs/stx/stx/libview/Depth8Image.st,v 1.28 1995-09-07 12:29:01 claus Exp $
'!

!Depth8Image class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libview/Depth8Image.st,v 1.28 1995-09-07 12:29:01 claus Exp $
"
!

documentation
"
    this class represents 256-color (8 bit / pixel) images (palette, greyscale ...).
    It mainly consists of methods already implemented in Image,
    reimplemented here for more performance.
"
! !

!Depth8Image class methodsFor:'queries'!

imageDepth
    ^ 8
! !

!Depth8Image methodsFor:'queries'!

bitsPerPixel
    "return the number of bits per pixel"

    ^ 8
!

bitsPerRow
    "return the number of bits in one scanline of the image"

    ^  width * 8
!

bitsPerSample
    "return the number of bits per sample.
     The return value is an array of bits-per-plane."

    ^ #(8)
!

bytesPerRow
    "return the number of bytes in one scanline of the image"

    ^ width
!

samplesPerPixel
    "return the number of samples per pixel in the image."

    ^ 1
!

usedValues
    "return a collection of color values used in the receiver."

    ^ bytes usedValues
! !

!Depth8Image methodsFor:'accessing'!

atX:x y:y
    "retrieve a pixel at x/y; return a color.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel"

    |value "{ Class: SmallInteger }"
     index "{ Class: SmallInteger }"|

    index := (width * y) + 1 + x.
    value := bytes at:index.

    photometric == #whiteIs0 ifTrue:[
	^ Color grey:100 - (100 / 255 * value)
    ].
    photometric == #blackIs0 ifTrue:[
	^ Color grey:(100 / 255 * value)
    ].
    photometric ~~ #palette ifTrue:[
	self error:'format not supported'.
	^ nil
    ].
    index := value + 1.
    ^ colorMap at:index
!

valueAtX:x y:y
    "retrieve a pixel at x/y; return a pixelValue.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel"

    |index "{ Class: SmallInteger }"|

    index := (width * y) + 1 + x.
    ^ bytes at:index.
!

atX:x y:y putValue:aPixelValue
    "set the pixel at x/y to aPixelValue.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel"

    |index "{ Class: SmallInteger }"|

    index := (width * y) + 1 + x.
    bytes at:index put:aPixelValue.
! !

!Depth8Image methodsFor:'enumerating'!

valueAtY:y from:xLow to:xHigh do:aBlock
    "perform aBlock for each pixelValue from x1 to x2 in row y.
     The block is passed the pixelValue at each pixel.
     This method allows slighly faster processing of an
     image than using valueAtX:y:, since some processing can be
     avoided when going from pixel to pixel. However, for
     real image processing, specialized methods should be written."

    |srcIndex   "{ Class: SmallInteger }"
     pixelValue "{ Class: SmallInteger }"
     x1         "{ Class: SmallInteger }"
     x2         "{ Class: SmallInteger }"|

    x1 := xLow.
    x2 := xHigh.
    srcIndex := (width * y) + 1 + x1.

    x1 to:x2 do:[:x |
	pixelValue := bytes at:srcIndex.
	srcIndex := srcIndex + 1.
	aBlock value:x value:pixelValue 
    ]
!

atY:y from:xLow to:xHigh do:aBlock
    "perform aBlock for each pixel from x1 to x2 in row y.
     The block is passed the color at each pixel.
     This method allows slighly faster processing of an
     image than using atX:y:, since some processing can be
     avoided when going from pixel to pixel. However, for
     real image processing, specialized methods should be written."

    |srcIndex "{ Class: SmallInteger }"
     index    "{ Class: SmallInteger }"
     value    "{ Class: SmallInteger }"
     x1       "{ Class: SmallInteger }"
     x2       "{ Class: SmallInteger }"
     color colors last|

    photometric == #palette ifTrue:[
	colors := colorMap.
    ] ifFalse:[
	colors := Array new:256.
    ].

    x1 := xLow.
    x2 := xHigh.
    srcIndex := (width * y) + 1 + x1.

    x1 to:x2 do:[:x |
	value := bytes at:srcIndex.
	srcIndex := srcIndex + 1.
	value ~~ last ifTrue:[
	    last := value.
	    index := value + 1.
	    color := colors at:index.
	    color isNil ifTrue:[
		photometric == #whiteIs0 ifTrue:[
		    color := (Color grey:100 - (100 * value / 255))
		] ifFalse:[
		    photometric == #blackIs0 ifTrue:[
			color := (Color grey:(100 * value / 255))
		    ] ifFalse:[
			photometric ~~ #palette ifTrue:[
			    self error:'format not supported'.
			    ^ nil
			]
		    ]
		].
		colors at:index put:color
	    ]
	].
	aBlock value:x value:color
    ]
! !

!Depth8Image methodsFor:'private'!

dither1PlaneUsingMap:map on:aDevice
    "a helper for dithering palette and greyscale images"

    |f
     patterns formBytes
     pixel0bytes pixel1bytes
     clr ditherPattern
     nColors       "{Class: SmallInteger }"
     w             "{Class: SmallInteger }"
     h             "{Class: SmallInteger }"|

    nColors := map size.
    w := width.
    h := height.

    formBytes := ByteArray uninitializedNew:(w + 7 // 8) * h.
    patterns := Array new:nColors.
    pixel0bytes := ByteArray uninitializedNew:nColors.
    pixel1bytes := ByteArray uninitializedNew:nColors.

    "
     extract dither patterns and values to use for 1/0 bits
     in those from the dithercolors
    "
    1 to:nColors do:[:i |
	clr := (map at:i) on:aDevice.
	ditherPattern := clr ditherForm.

	ditherPattern isNil ifTrue:[
	    patterns at:i put:#[2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111].
	    pixel0bytes at:i put:clr colorId.
	    pixel1bytes at:i put:clr colorId
	] ifFalse:[
	    patterns at:i put:(ditherPattern bits).
	    pixel0bytes at:i put:(ditherPattern colorMap at:1) colorId.
	    pixel1bytes at:i put:(ditherPattern colorMap at:2) colorId.
	].
    ].

%{
    unsigned char *_srcP, *_dstP;
    OBJ _patternBytes;
    unsigned char _mask = 0x80;
    unsigned char _outBits = 0;
    unsigned char _last, _v, _patternBits, _p0, _p1;
    int _h, _w;
    int _patternOffset = 0;

    _srcP = _ByteArrayInstPtr(_INST(bytes))->ba_element;
    _dstP = _ByteArrayInstPtr(formBytes)->ba_element;
    for (_h = _intVal(h); _h; _h--) {
	_last = -1;
	for (_w = _intVal(w); _w; _w--) {
	    _v = *_srcP++;
	    if (_v != _last) {
		_patternBytes = _ArrayInstPtr(patterns)->a_element[_v];
		if (__isByteArray(_patternBytes)) {
		    _patternBits = _ByteArrayInstPtr(_patternBytes)->ba_element[_patternOffset];
		} else if (__isArray(_patternBytes)) {
		    _patternBits = _intVal(_ArrayInstPtr(_patternBytes)->a_element[_patternOffset]);
		}
		_p0 = _ByteArrayInstPtr(pixel0bytes)->ba_element[_v];
		_p1 = _ByteArrayInstPtr(pixel1bytes)->ba_element[_v];
		_last = _v;
	    }
	    _outBits <<= 1;
	    if (_patternBits & _mask)
		_outBits |= _p1;
	    else
		_outBits |= _p0;
	    _mask >>= 1;
	    if (_mask == 0) {
		_mask = 0x80;
		*_dstP++ = _outBits;
		_outBits = 0;
	    }
	}
	if (_mask != 0x80) {
	    while (_mask != 0) {
		_outBits <<= 1;
		_mask >>= 1;
	    }
	    *_dstP++ = _outBits;
	    _mask = 0x80;
	    _outBits = 0;
	}
	_patternOffset++;
	if (_patternOffset == 8)
	    _patternOffset = 0;
    }
%}
.
    f := Form width:w height:h fromArray:formBytes.
    ^ f
!

dither2PlaneUsingMap:map on:aDevice
    "a helper for dithering palette and greyscale images"

    |f
     patterns formBytes
     pixel0bytes pixel1bytes
     clr ditherPattern
     nColors       "{Class: SmallInteger }"
     w             "{Class: SmallInteger }"
     h             "{Class: SmallInteger }"|

    nColors := map size.
    w := width.
    h := height.

    formBytes := ByteArray uninitializedNew:(w * 2 + 7 // 8) * h.
    patterns := Array new:nColors.
    pixel0bytes := ByteArray uninitializedNew:nColors.
    pixel1bytes := ByteArray uninitializedNew:nColors.

    "extract dither patterns and values to use for 1/0 bits
     in those from the dithercolors"

    1 to:nColors do:[:i |
	clr := (map at:i) on:aDevice.
	ditherPattern := clr ditherForm.

	ditherPattern isNil ifTrue:[
	    patterns at:i put:#[2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111
				2r11111111].
	    pixel0bytes at:i put:clr colorId.
	    pixel1bytes at:i put:clr colorId
	] ifFalse:[
	    patterns at:i put:(ditherPattern bits).
	    pixel0bytes at:i put:(ditherPattern colorMap at:1) colorId.
	    pixel1bytes at:i put:(ditherPattern colorMap at:2) colorId.
	].
    ].

%{
    unsigned char *_srcP, *_dstP;
    OBJ _patternBytes;
    unsigned char _mask = 0x80;
    unsigned char _outBits = 0;
    unsigned char _last, _v, _patternBits, _p0, _p1;
    int _h, _w;
    int _patternOffset = 0;
    int _outCount;

    _srcP = _ByteArrayInstPtr(_INST(bytes))->ba_element;
    _dstP = _ByteArrayInstPtr(formBytes)->ba_element;
    for (_h = _intVal(h); _h; _h--) {
	_last = -1;
	_outCount = 0;
	for (_w = _intVal(w); _w; _w--) {
	    _v = *_srcP++;
	    if (_v != _last) {
		_patternBytes = _ArrayInstPtr(patterns)->a_element[_v];
		if (__isByteArray(_patternBytes)) {
		    _patternBits = _ByteArrayInstPtr(_patternBytes)->ba_element[_patternOffset];
		} else if (__isArray(_patternBytes)) {
		    _patternBits = _intVal(_ArrayInstPtr(_patternBytes)->a_element[_patternOffset]);
		}
		_p0 = _ByteArrayInstPtr(pixel0bytes)->ba_element[_v];
		_p1 = _ByteArrayInstPtr(pixel1bytes)->ba_element[_v];
		_last = _v;
	    }
	    _outBits <<= 2;
	    if (_patternBits & _mask)
		_outBits |= _p1;
	    else
		_outBits |= _p0;
	    _mask >>= 1;
	    _outCount++;
	    if (_outCount == 4) {
		*_dstP++ = _outBits;
		_outCount = 0;
		if (_mask == 0) {
		    _mask = 0x80;
		}
	    }
	}
	if (_outCount) {
	    do {
		_outBits <<= 2;
	    } while (++_outCount != 4);
	    *_dstP++ = _outBits;
	}
	_mask = 0x80;
	_outBits = 0;
	_patternOffset++;
	if (_patternOffset == 8)
	    _patternOffset = 0;
    }
%}
.
    f := Form width:w height:h depth:2.
    f initGC.
    f device drawBits:formBytes depth:2 width:w height:h x:0 y:0
		 into:f id x:0 y:0 width:w height:h with:f gcId.
    ^ f
! !

!Depth8Image methodsFor:'converting palette images'!

paletteImageAsMonoFormOn:aDevice
    "return a 1-bit monoForm from the palette picture -
     the result is a thresholded form, with white for
     brightness values above 50%, black below"

    |monoBits f
     map
     failed
     w        "{ Class: SmallInteger }"
     h        "{ Class: SmallInteger }"
"/     v        "{ Class: SmallInteger }"
"/     bitCount "{ Class: SmallInteger }"
"/     bits     "{ Class: SmallInteger }"
"/     srcIndex "{ Class: SmallInteger }"
"/     dstIndex "{ Class: SmallInteger }"
     mapSize  "{ Class: SmallInteger }"|

    w := width.
    h := height.
    monoBits := ByteArray uninitializedNew:(((w + 7) // 8) * h).

    map := ByteArray uninitializedNew:256.
    mapSize := colorMap size.

    "
     map entries: 0 for dark entries, 1 for bright entries
    "
    1 to:mapSize do:[:i |
	map at:i put:(colorMap at:i) brightness rounded
    ].

    failed := true.
%{
    register unsigned char *srcPtr, *dstPtr, *mapPtr;
    register __v, __bits, __bitCount;
    register j;
    register i;
    extern OBJ ByteArray;

    if (__isByteArray(_INST(bytes))
     && __isByteArray(map)
     && __isByteArray(monoBits)) {
	failed = false;
	srcPtr = _ByteArrayInstPtr(_INST(bytes))->ba_element;
	dstPtr = _ByteArrayInstPtr(monoBits)->ba_element;
	mapPtr = _ByteArrayInstPtr(map)->ba_element;
	for (i=_intVal(h); i>0; i--) {
	    __bitCount = 0;
	    __bits = 0;
	    for (j=_intVal(w); j>0; j--) {
		__v = mapPtr[*srcPtr++];
		__bits = (__bits<<1) | __v; 
		__bitCount++;
		if (__bitCount == 8) {
		    *dstPtr++ = __bits;
		    __bits = 0;
		    __bitCount = 0;
		}
	    }
	    if (__bitCount != 0) {
		*dstPtr++ = __bits;
	    }
	}
    }
%}
.
    failed ifTrue:[
"/
"/ the above code is equivalent to:
"/
"/        srcIndex := 1.
"/        dstIndex := 1.
"/        1 to:h do:[:row |
"/
"/            bitCount := 0.
"/            bits := 0.
"/            1 to:w do:[:col |
"/                v := bytes at:srcIndex.
"/                srcIndex := srcIndex + 1.
"/                v := map at:(v + 1).
"/                bits := (bits bitShift:1) bitOr:v.
"/                bitCount := bitCount + 1.
"/                (bitCount == 8) ifTrue:[
"/                    monoBits at:dstIndex put:bits.
"/                    dstIndex := dstIndex + 1.
"/                    bits := 0.
"/                    bitCount := 0
"/                ]
"/            ].
"/            (bitCount ~~ 0) ifTrue:[
"/                monoBits at:dstIndex put:bits.
"/                dstIndex := dstIndex + 1
"/            ]
"/        ]
"/
"/ we dont need the fall-back code; so trigger an error
	self primitiveFailed.
	^ nil
    ].

    f := Form width:w height:h depth:1 on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.
"/    (aDevice blackpixel == 0) ifFalse:[
"/        "have to invert bits"
"/        f function:#copyInverted
"/    ].
    aDevice drawBits:monoBits depth:1 width:w height:h
		   x:0 y:0
		into:(f id) x:0 y:0 width:w height:h with:(f gcId).
    ^ f
!

paletteImageAs2PlaneFormOn:aDevice
    "return a 2-bit greyForm from the palette picture -
     the result is a thresholded form, with white/lightGrey/darkGrey
     and black for brightness values 100..75, 75..50, 50..25 and 25..0 %"

    |twoPlaneBits f
     map failed
     w        "{ Class: SmallInteger }"
     h        "{ Class: SmallInteger }"
"/     v        "{ Class: SmallInteger }"
"/     bitCount "{ Class: SmallInteger }"
"/     bits     "{ Class: SmallInteger }"
"/     srcIndex "{ Class: SmallInteger }"
"/     dstIndex "{ Class: SmallInteger }"
     mapSize  "{ Class: SmallInteger }" |

    w := width.
    h := height.
    twoPlaneBits := ByteArray uninitializedNew:(((w * 2 + 7) // 8) * h).

    map := ByteArray uninitializedNew:256.
    mapSize := colorMap size.

    "
     map entries: 0 .. 3 give brightness in 4 thresholded steps
    "
    1 to:mapSize do:[:i |
	map at:i put:(colorMap at:i) brightness * 3 rounded
    ].

    failed := true.
%{
    register unsigned char *srcPtr, *dstPtr, *mapPtr;
    register __v, __bits, __bitCount;
    register j;
    register i;
    extern OBJ ByteArray;

    if ((__isByteArray(_INST(bytes)))
     && (__isByteArray(map))
     && (__isByteArray(twoPlaneBits))) {
	failed = false;
	srcPtr = _ByteArrayInstPtr(_INST(bytes))->ba_element;
	dstPtr = _ByteArrayInstPtr(twoPlaneBits)->ba_element;
	mapPtr = _ByteArrayInstPtr(map)->ba_element;
	for (i=_intVal(h); i>0; i--) {
	    __bitCount = 0;
	    __bits = 0;
	    for (j=_intVal(w); j>0; j--) {
		__v = mapPtr[*srcPtr++];
		__bits = (__bits<<2) | __v; 
		__bitCount++;
		if (__bitCount == 4) {
		    *dstPtr++ = __bits;
		    __bits = 0;
		    __bitCount = 0;
		}
	    }
	    if (__bitCount != 0) {
		*dstPtr++ = __bits;
	    }
	}
    }
%}.
    failed ifTrue:[
"/
"/ the above code is equivalent to:
"/
"/        srcIndex := 1.
"/        dstIndex := 1.
"/        1 to:h do:[:row |
"/            bitCount := 0.
"/            bits := 0.
"/            1 to:w do:[:col |
"/                v := bytes at:srcIndex.
"/                srcIndex := srcIndex + 1.
"/                v := map at:(v + 1).
"/                bits := (bits bitShift:2) bitOr:v.
"/                bitCount := bitCount + 1.
"/                (bitCount == 4) ifTrue:[
"/                    twoPlaneBits at:dstIndex put:bits.
"/                    dstIndex := dstIndex + 1.
"/                    bits := 0.
"/                    bitCount := 0
"/                ]
"/            ].
"/            (bitCount ~~ 0) ifTrue:[
"/                twoPlaneBits at:dstIndex put:bits.
"/                dstIndex := dstIndex + 1
"/            ]
"/        ]
"/
	self primitiveFailed.
	^ nil
    ].

    f := Form width:w height:h depth:2 on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.
    (aDevice blackpixel == 0) ifFalse:[
	"have to invert bits"
	f function:#copyInverted
    ].
    aDevice drawBits:twoPlaneBits depth:2 width:w height:h
		   x:0 y:0
		into:(f id) x:0 y:0 width:w height:h with:(f gcId).
    ^ f
!

paletteImageAsDitheredPseudoFormOn:aDevice
    "return a dithered pseudoForm from the palette picture. Depend
     on dither colors being preallocated (see Color>>getColors*)"

    ^ self paletteImageAsDitheredPseudoFormOn:aDevice 
	   colors:Color fixColors 
	   nRed:Color numFixRed
	   nGreen:Color numFixGreen
	   nBlue:Color numFixBlue
!

paletteImageAsDitheredPseudoFormOn:aDevice colors:fixColors nRed:nRed nGreen:nGreen nBlue:nBlue 
    "return a dithered pseudoForm from the palette picture. 
     Use the colors in the fixColors array, which must be fixR x fixG x fixB
     colors assigned to aDevice, such as the preallocated colors of the
     Color class. 
     By passing the ditherColors as extra array, this method can
     also be used to dither an 8bit image into a smaller number of colors,
     for example to create Depth4Images."

    "the code below is a q&d hack - it needs a rewrite to use a
     floyd-steinberg dither (currently, the error is only forwarded
     to the next pixel on the right, which seems ok for photograph-like
     images, but leads to more artifacts with equal colored areas).
    "

    |pseudoBits f has8BitImage deviceDepth
     rgbBytes
     w     "{Class: SmallInteger }"
     h     "{Class: SmallInteger }"
     index "{Class: SmallInteger }"
     fixR  "{Class: SmallInteger }"
     fixG  "{Class: SmallInteger }"
     fixB  "{Class: SmallInteger }"
     fixIds failed map colorMapSize
     error|

    aDevice ~~ Display ifTrue:[^ nil].

    fixR := nRed.
    fixR == 0 ifTrue:[ ^ nil].
    fixG := nGreen.
    fixG == 0 ifTrue:[ ^ nil].
    fixB := nBlue.
    fixB == 0 ifTrue:[ ^ nil].
    "/ simple check
    (fixR * fixG * fixB) ~~ fixColors size ifTrue:[
	self error:'invalid color array passed'.
	^ nil
    ].
    fixIds := (fixColors asArray collect:[:clr | clr colorId]) asByteArray.

    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
	has8BitImage := true.
    ] ifFalse:[
	has8BitImage := false.
	aDevice supportedImageFormats do:[:fmt |
	    (fmt at:2) == 8 ifTrue:[
		has8BitImage := true.
	    ]
	]
    ].
    has8BitImage ifFalse:[^ nil].

    'D8IMAGE: dithering ...' infoPrintNL.

    "
     collect color components as integer values
     (code below uses components percent * 2.55 asInteger everywhere, to avoid
      float arithmetic, rounding etc. Thus, the range is 0..255 here)
    "
    colorMapSize := colorMap size.
    rgbBytes := ByteArray uninitializedNew:colorMapSize * 3.

    index := 1.
    1 to:colorMapSize do:[:i |
	|clr|

	clr := colorMap at:i.
	rgbBytes at:index put:(clr red * 2.55) asInteger.
	rgbBytes at:index+1 put:(clr green * 2.55) asInteger.
	rgbBytes at:index+2 put:(clr blue * 2.55) asInteger.
	index := index + 3.
    ].

    pseudoBits := ByteArray uninitializedNew:(width * height).

    w := width + 2.
    error := ByteArray new:w*(3*2).

    w := width.
    h := height.

    failed := true.

%{
/*
 * #   define FLOYD_STEINBERG
 */
    int __x, __y;
    int __eR, __eG, __eB;
    unsigned char *srcP, *dstP;
    unsigned char *rgbP;
    unsigned char *idP;
    short *errP, *eP;
    int __fR, __fG, __fB;
    int iR, iG, iB;
    int idx;

    if (__isByteArray(_INST(bytes))
     && __isByteArray(pseudoBits)
     && __isByteArray(rgbBytes)
     && __isByteArray(fixIds)
     && __isByteArray(error)
     && __bothSmallInteger(fixR, fixG)
     && __isSmallInteger(fixB)) {
	failed = false;

	srcP = _ByteArrayInstPtr(_INST(bytes))->ba_element;
	dstP = _ByteArrayInstPtr(pseudoBits)->ba_element;
	rgbP = _ByteArrayInstPtr(rgbBytes)->ba_element;
	idP = _ByteArrayInstPtr(fixIds)->ba_element;
	errP = (short *) _ByteArrayInstPtr(error)->ba_element;
	__fR = _intVal(fixR)-1;
	__fG = _intVal(fixG)-1;
	__fB = _intVal(fixB)-1;

#ifdef FLOYD_STEINBERG
	/*
	 * clear error accumulator
	 */
	eP = errP;
	for (__x=_intVal(w)+1; __x>=0; __x--) {
	    eP[0] = 0; eP[1] = 0; eP[2] = 0;
	    eP += 3;
	}
#endif
	for (__y=_intVal(h); __y>0; __y--) {
	    __eR = __eG = __eB = 0;
#ifdef FLOYD_STEINBERG
	    eP = &(errP[3]);
	    __eR += eP[0];
	    __eG += eP[1];
	    __eB += eP[2];
#endif

	    for (__x=_intVal(w); __x>0; __x--) {
		int __want;
		int pix;
		int __wantR, __wantG, __wantB;
		int idx;
		int tR, tG, tB;
		int nR, nG, nB;

		pix = *srcP++;

		/*
		 * wR, wG and wB is the wanted r/g/b value;
		 * compute the index into the dId table ..
		 * values: 0..255; scale to 0..fR-1, 0..fG-1, 0..fB-1
		 *
		 * bad kludge: knows how to index into FixColor table
		 */
		idx = pix+pix+pix;  /* pix * 3 */

		__wantR = rgbP[idx]   + __eR;
		__wantG = rgbP[idx+1] + __eG;
		__wantB = rgbP[idx+2] + __eB;

		if (__wantR > 255) __want = 255;
		else if (__wantR < 0) __want = 0;
		else __want = __wantR;

		iR = (__want * __fR + 128) / 255; /* red index rounded */
		idx = iR * (__fG+1);

		if (__wantG > 255) __want = 255;
		else if (__wantG < 0) __want = 0;
		else __want = __wantG;

		iG = (__want * __fG + 128) / 255; /* green index rounded */
		idx = (idx + iG) * (__fB+1);

		if (__wantB > 255) __want = 255;
		else if (__wantB < 0) __want = 0;
		else __want = __wantB;

		iB = (__want * __fB + 128) / 255; /* blue index rounded */
		idx = idx + iB;

		/*
		 * store the corresponding dither colors colorId
		 */
		*dstP++ = idP[idx];

		/*
		 * the new error:
		 */
		__eR = __wantR - (iR * 255 / __fR); 
		__eG = __wantG - (iG * 255 / __fG); 
		__eB = __wantB - (iB * 255 / __fB); 
#ifdef FLOYD_STEINBERG
		/*
		 * distribute the error
		 */
		tR = __eR >> 4;  /* 16th of error */
		tG = __eG >> 4;
		tB = __eB >> 4;

		nR = eP[3] + (tR * 7);/* from accu: error for (x+1 / y) */
		nG = eP[4] + (tG * 7);/* plus 7/16'th of this error */
		nB = eP[5] + (tB * 7);

		eP[0] = tR*5;         /* 5/16th for (x / y+1) */
		eP[1] = tG*5;
		eP[2] = tB*5;

		eP[-3] = tR*3;        /* 3/16th for (x-1 / y+1) */
		eP[-2] = tG*3;
		eP[-1] = tB*3;

		eP[3] = __eR - (tR*15);  /* 1/16th for (x+1 / y+1) */
		eP[4] = __eG - (tG*15);
		eP[5] = __eB - (tB*15);

		__eR = nR;
		__eG = nG;
		__eB = nB;

		eP += 3;
#endif
	    }
	}
    }
%}.
    failed ifTrue:[
	self primitiveFailed.
	^ nil
    ].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].
    "/
    "/ have to create a funny colorMap, where
    "/ color at:index == color colorId:index
    "/
    map := Array new:256.
    fixColors do:[:clr |
	map at:clr colorId + 1 put:clr
    ].
    f colorMap:map. 
    f initGC.
    f bits:pseudoBits.
    aDevice drawBits:pseudoBits bitsPerPixel:8 depth:deviceDepth  
	       width:width height:height
		   x:0 y:0
		into:(f id) x:0 y:0 
	       width:width height:height with:(f gcId).
    ^ f

    "
     example: 
	color reduction from Depth8 to Depth4 (dithering) can be done by:

     |img8 reducedImg8 img4 map form|

     map := #( 
		  (0     0   0)
		  (0     0 100)
		  (0    50   0)
		  (0    50 100)
		  (0   100   0)
		  (0   100 100)
		  (100   0   0)
		  (100   0 100)
		  (100  50   0)
		  (100  50 100)
		  (100 100   0)
		  (100 100 100)) collect:[:rgb | (Color red:(rgb at:1)
						      green:(rgb at:2)
						       blue:(rgb at:3)) on:Display].

     img8 := Image fromFile:'bitmaps/bf.im8'.
     form := img8 paletteImageAsDitheredPseudoFormOn:Display 
		      colors:map 
			nRed:2
		      nGreen:3
		       nBlue:2.
     img8 := Depth8Image fromForm:form.    'dithered version of original image'.
     img4 := Depth4Image fromImage:img8.
    "
!

paletteImageAsPseudoFormOn:aDevice
    "return a pseudoForm from the palette picture. The main work is
     in color reduction, when not all colors can be aquired."

    |pseudoBits f gcRound has8BitImage deviceDepth
     imgMap newImage pxl
     usedColors usageCounts maxIndex map
     fit scale lastOK error 
     div 
     bitsPerRGB "{Class: SmallInteger }"
     shift      "{Class: SmallInteger }"
     m          "{Class: SmallInteger }" |

    Color fixColors notNil ifTrue:[
	f := self paletteImageAsDitheredPseudoFormOn:aDevice.
	f notNil ifTrue:[^ f].
    ].

    "find used colors"

    usedColors := bytes usedValues.    "gets us an array filled with used values"
				       "(could use bytes asBag)"
    maxIndex := usedColors max + 1.

    usedColors size > 20 ifTrue:[
	('D8IMAGE: allocating ' , usedColors size printString , ' colors ...') infoPrintNL.
    ].

    "sort by usage"
    usageCounts := bytes usageCounts.
    usageCounts := usedColors asArray collect:[:clr | usageCounts at:(clr + 1)].
    usageCounts sort:[:a :b | a > b] with:usedColors.

    "allocate the colors (in order of usage count)"

    imgMap := Array new:maxIndex.

    "
     first, try to get the exact colors ...
    "
    bitsPerRGB := aDevice bitsPerRGB.
    shift := (8 - bitsPerRGB) negated.
    m := (1 bitShift:bitsPerRGB) - 1.
    div := m asFloat.

    fit := true.
    scale := 100.0 / div.       "to scale 0..255 into 0.0 .. 100.0"
    lastOK := 0.
    gcRound := 0.

    usedColors do:[:aColorIndex |
	|devColor color
	 r        "{Class: SmallInteger }"
	 g        "{Class: SmallInteger }"
	 b        "{Class: SmallInteger }"
	 mapIndex "{Class: SmallInteger }"
	 rMask    "{Class: SmallInteger }"
	 gMask    "{Class: SmallInteger }"
	 bMask    "{Class: SmallInteger }"|

	fit ifTrue:[
	    gMask := bMask := rMask := m.

	    mapIndex := aColorIndex + 1.
	    color := colorMap at:mapIndex.
	    color colorId notNil ifTrue:[
		"wow - an immediate hit"
		devColor := color
	    ] ifFalse:[
		devColor := color exactOn:aDevice.
		devColor isNil ifTrue:[
		    "
		     could not allocate color - on the first round, do a GC to flush 
		     unused colors - this may help if some colors where locked by 
		     already free images.
		    "
		    gcRound == 0 ifTrue:[
			ObjectMemory scavenge; finalize.
			devColor := color exactOn:aDevice.
			gcRound := 1
		    ].
		    devColor isNil ifTrue:[
			gcRound == 1 ifTrue:[
			    CollectGarbageWhenRunningOutOfColors ifTrue:[
				'D8IMAGE: force GC for possible color reclamation.' infoPrintNL.
				ObjectMemory incrementalGC; finalize.
				devColor := color exactOn:aDevice.
			    ].    
			    gcRound := 2
			]
		    ]
		].
	    ].
	    (devColor notNil and:[devColor colorId notNil]) ifTrue:[
		imgMap at:mapIndex put:devColor.
		lastOK := lastOK + 1.
	    ] ifFalse:[
		fit := false
	    ]
	]
    ].

    fit ifFalse:[
"/        |tree|
"/
"/        "/ first, create an octTree containing colors which we got ...
"/        tree := ColorOctree new.
"/
"/        usedColors from:1 to:lastOK do:[:aColorIndex |
"/            tree insert:(colorMap at:aColorIndex + 1).
"/        ].
"/
"/        "/ then, remap remaining, using nearest from those already allocated
"/
"/        usedColors from:(lastOK+1) to:(usedColors size) do:[:aColorIndex |      
"/            |mapIndex|
"/
"/            mapIndex := aColorIndex + 1.
"/            imgMap at:mapIndex put:(tree findBest:(colorMap at:mapIndex))
"/        ].

	"
	 again, this time allow wrong colors (loop while increasing allowed error)
	"
	error := 10.
	[fit] whileFalse:[
	    fit := true.
	    usedColors from:(lastOK+1) to:(usedColors size) do:[:aColorIndex |
		|devColor color
		 r        "{Class: SmallInteger }"
		 g        "{Class: SmallInteger }"
		 b        "{Class: SmallInteger }"
		 mapIndex "{Class: SmallInteger }"
		 rMask    "{Class: SmallInteger }"
		 gMask    "{Class: SmallInteger }"
		 bMask    "{Class: SmallInteger }"|

		fit ifTrue:[
		    gMask := bMask := rMask := m.

		    mapIndex := aColorIndex + 1.
		    color := colorMap at:mapIndex.
		    r := (color red * 255 / 100) rounded.
		    g := (color green * 255 / 100) rounded.
		    b := (color blue * 255 / 100) rounded.

		    color := Color red:((r bitShift:shift) bitAnd:rMask) * scale
				 green:((g bitShift:shift) bitAnd:gMask) * scale
				  blue:((b bitShift:shift) bitAnd:bMask) * scale.

		    color colorId notNil ifTrue:[
			"wow - an immediate hit"
			devColor := color
		    ] ifFalse:[
			devColor := color nearestOn:aDevice error:error.
			devColor isNil ifTrue:[
			    "
			     no free color - on the first round, do a GC to flush unused
			     colors - this may help if some colors where locked by already
			     free images.
			    "
			    gcRound == 0 ifTrue:[
				ObjectMemory scavenge; finalize.
				devColor := color nearestOn:aDevice error:error.
				gcRound := 1
			    ].
			    devColor isNil ifTrue:[
				gcRound == 1 ifTrue:[
				    'D8IMAGE: force GC for possible color reclamation.' infoPrintNL.
				    ObjectMemory incrementalGC; finalize.
				    devColor := color nearestOn:aDevice error:error.
				    gcRound := 2
				]
			    ]
			].
		    ].
		    (devColor notNil and:[devColor colorId notNil]) ifTrue:[
			imgMap at:mapIndex put:devColor.
			lastOK := lastOK + 1.
		    ] ifFalse:[
			fit := false
		    ]
		].
	    ].
	    error := error * 2.
	    error > 1000 ifTrue:[
		"
		 break out, if the error becomes too big.
		"
		'D8IMAGE: hard color allocation problem - revert to b&w' infoPrintNL.
		"
		 map to b&w as a last fallback.
		 (should really do a dither here)
		"
		usedColors from:(lastOK+1) to:(usedColors size) do:[:aColorIndex |
		    |color
		     mapIndex "{ Class: SmallInteger }"|

		    mapIndex := aColorIndex + 1.
		    color := colorMap at:mapIndex.
		    color brightness > 0.5 ifTrue:[
			color := Color white.
		    ] ifFalse:[
			color := Color black.
		    ].
		    imgMap at:mapIndex put:(color on:aDevice).
		].
		fit := true.
	    ]
	].

	error > 100 ifTrue:[
	    'D8IMAGE: not enough colors for a reasonable image' infoPrintNL
	] ifFalse:[
	    'D8IMAGE: not enough colors for exact picture' infoPrintNL.
	]
    ].

    "
     create translation map (from image colors to allocated colorIds)
    "
    map := ByteArray new:256.
    1 to:imgMap size do:[:i |
	|clr|

	(clr := imgMap at:i) notNil ifTrue:[
	    map at:i put:clr colorId
	]
    ].

    "
     does the device support 8-bit images ?
    "
    deviceDepth := aDevice depth.
    deviceDepth == 8 ifTrue:[
	has8BitImage := true.
    ] ifFalse:[
	has8BitImage := false.
	aDevice supportedImageFormats do:[:fmt |
	    (fmt at:2) == 8 ifTrue:[
		has8BitImage := true.
	    ]
	]
    ].

    "
     finally, create a form on the device and copy (& translate)
     the pixel values
    "
    has8BitImage ifTrue:[
	pseudoBits := ByteArray uninitializedNew:(width * height).

	bytes expandPixels:8         "xlate only"
		    width:width 
		   height:height
		     into:pseudoBits
		  mapping:map.

	map := nil.

	f := Form width:width height:height depth:deviceDepth on:aDevice.
	f isNil ifTrue:[^ nil].
	f colorMap:imgMap. 
	f initGC.
	aDevice drawBits:pseudoBits bitsPerPixel:8 depth:deviceDepth  
		   width:width height:height
		       x:0 y:0
		    into:(f id) x:0 y:0 
		   width:width height:height with:(f gcId).
	^ f
    ].

    "
     slow fall back: convert into appropriate depth image,
     by looping over each pixel individually
    "
    newImage := (Image implementorForDepth:deviceDepth) new.
    newImage width:width.
    newImage height:height.
    newImage bits:(ByteArray uninitializedNew:(height * newImage bytesPerRow)).

    0 to:height-1 do:[:row |
	0 to:width-1 do:[:col |
	    pxl := self valueAtX:col y:row.
	    newImage atX:col y:row putValue:(map at:pxl)
	]
    ].

    f := Form width:width height:height depth:deviceDepth on:aDevice.
    f isNil ifTrue:[^ nil].
    f colorMap:imgMap. 
    f initGC.

    aDevice drawBits:(newImage bits) depth:deviceDepth width:width height:height
		   x:0 y:0
		into:(f id) x:0 y:0 
	       width:width height:height with:(f gcId).

    ^ f
!

paletteImageAsGreyFormOn:aDevice
    "return an 8-bit greyForm from the 8-bit palette picture;
     only a translation has to be done"

    |greyBits f map
     mapSize "{ Class: SmallInteger }"|

    greyBits := ByteArray uninitializedNew:(width * height).

    map := ByteArray uninitializedNew:256.
    mapSize := colorMap size.

    1 to:mapSize do:[:i |
	map at:i put:((colorMap at:i) brightness * 255) rounded
    ].

    bytes expandPixels:8         "xlate only"
		width:width 
	       height:height
		 into:greyBits
	      mapping:map.

    f := Form width:width height:height depth:8 on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.
    aDevice drawBits:greyBits depth:8 width:width height:height
		       x:0 y:0
		    into:(f id) x:0 y:0 
		   width:width height:height with:(f gcId).
    ^ f
!

paletteImageAsPatteredGreyFormOn:aDevice
    "return a dithered greyForm from the 8-bit palette picture.
     works for any destination depth - but is very slow for some."

    |f 
     map run last ditherColors first delta
     clr depth grey
     nDither       "{Class: SmallInteger }"
     nColors       "{Class: SmallInteger }"
     w             "{Class: SmallInteger }"
     h             "{Class: SmallInteger }"
     v             "{Class: SmallInteger }"
     srcIndex      "{Class: SmallInteger }"|


    Transcript showCr:'dithering ..'. Transcript endEntry.

    nDither := NumberOfDitherColors.
    ditherColors := Array new:nDither.

    first := (100 / nDither / 2).
    delta := 100 / nDither.
    0 to:nDither-1 do:[:i |
	ditherColors at:i+1 put:(Color grey:(i * delta + first)).
    ].

    nColors := colorMap size.
    map := Array new:nColors.
    1 to:nColors do:[:i |
	clr := colorMap at:i.
	grey := clr brightness.
	map at:i put:(ditherColors at:(v * (nDither - 1)) rounded)
    ].

    "tuning - code below is so slooow"
    "get the patterns, fill form bytes here"

    depth := aDevice depth.
    depth == 1 ifTrue:[
	^ self dither1PlaneUsingMap:map on:aDevice
    ].
    depth == 2 ifTrue:[
	^ self dither2PlaneUsingMap:map on:aDevice
    ].

    "draw each pixel using dither color (let others do the dithering)
     although the code is simple, its very slow"

    w := width.
    h := height.

    f := Form width:w height:h depth:depth on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.

    srcIndex := 1.
    1 to:h do:[:dstY |
	run := 0.
	last := nil.
	1 to:w do:[:dstX |
	    v := bytes at:srcIndex.
	    srcIndex := srcIndex + 1.
	    clr := map at:(v + 1).

	    clr == last ifTrue:[
		run := run + 1
	    ] ifFalse:[
		(run ~~ 0) ifTrue:[
		    f fillRectangleX:dstX-run-1 y:dstY-1 width:run height:1.
		].
		run := 1.
		f paint:clr.
		last := clr
	    ].
	].
	f fillRectangleX:w-run y:dstY-1 width:run height:1.
    ].
    ^ f
! !

!Depth8Image methodsFor:'converting greyscale images'!

greyImageAsMonoFormOn:aDevice
    "return a (thresholded) monochrome Form from the picture."

    |bytesPerRow
     bytesPerMonoRow monoData
     pixel       "{Class: SmallInteger }"
     byte        "{Class: SmallInteger }"
     mask        "{Class: SmallInteger }"
     srcIndex    "{Class: SmallInteger }"
     dstIndex    "{Class: SmallInteger }"
     nextSrc     "{Class: SmallInteger }"
     nextDst     "{Class: SmallInteger }"
     bitNumber   "{Class: SmallInteger }"
     w           "{Class: SmallInteger }"
     h           "{Class: SmallInteger }" |

    w := width.
    h := height.

    bytesPerRow := self bytesPerRow.
    bytesPerMonoRow := w // 8.
    ((w \\ 8) ~~ 0) ifTrue:[
	bytesPerMonoRow := bytesPerMonoRow + 1
    ].
    monoData := ByteArray uninitializedNew:(bytesPerMonoRow * h).

    "8 bit -> 1 bit extract; take most significant bit"

    srcIndex := 1.
    dstIndex := 1.
    1 to:h do:[:count |
	nextSrc := srcIndex + bytesPerRow.
	nextDst := dstIndex + bytesPerMonoRow.
	bitNumber := 1.
	mask := 2r10000000.
	[bitNumber <= w] whileTrue:[
	    byte := 0.
	    pixel := bytes at:srcIndex.
	    srcIndex := srcIndex + 1.
	    ((pixel bitAnd:2r10000000) ~~ 0) ifTrue:[
		byte := byte bitOr:mask
	    ].
	    mask := mask bitShift: -1.
	    (mask == 0) ifTrue:[
		monoData at:dstIndex put:byte.
		dstIndex := dstIndex + 1.
		byte := 0.
		mask := 2r10000000
	    ].
	    bitNumber := bitNumber + 1
	].
	(mask == 2r10000000) ifFalse:[
	    monoData at:dstIndex put:byte.
	    dstIndex := dstIndex + 1
	].
	srcIndex := nextSrc.
	dstIndex := nextDst
    ].

    ^ Form width:w height:h fromArray:monoData on:aDevice
!

greyImageAsPatternDitheredGreyFormOn:aDevice
    "return a dithered greyForm from the grey picture.
     Works for any destination depth.
     Used to reduce the number of grey levels."

    |f depth
     map pixel0bytes pixel1bytes clr ditherPattern
     last ditherColors nDither first delta patterns formBytes
     w             "{Class: SmallInteger }"
     h             "{Class: SmallInteger }"
     v             "{Class: SmallInteger }"
     run           "{Class: SmallInteger }"
     srcIndex      "{Class: SmallInteger }"
     dstIndex      "{Class: SmallInteger }"
     mask          "{Class: SmallInteger }"
     outBits       "{Class: SmallInteger }"
     patternOffset "{Class: SmallInteger }"
     patternBytes
     patternBits   "{Class: SmallInteger }"
     index         "{Class: SmallInteger }"
     p0            "{Class: SmallInteger }"
     p1            "{Class: SmallInteger }" |

    Transcript showCr:'dithering ..'. Transcript endEntry.

    nDither := NumberOfDitherColors.
    ditherColors := Array new:nDither.

    first := (100 / nDither / 2).
    delta := 100 / nDither.
    0 to:nDither-1 do:[:i |
	ditherColors at:i+1 put:(Color grey:(i * delta + first)).
    ].

    map := Array new:256.
    1 to:256 do:[:i |
	v := i - 1.
	v := (v * (nDither - 1) // 255) rounded.
	" v is now 0 .. nDither-1 "
	map at:i put:(ditherColors at:(v + 1))
    ].

    depth := aDevice depth.

    "tuning (general code is too slow)
     get the patterns, fill form bytes here"

    w := width.
    h := height.

    depth == 1 ifTrue:[
	^ self dither1PlaneUsingMap:map on:aDevice.

        
"/      formBytes := ByteArray uninitializedNew:(w + 7 // 8) * h.
"/      patterns := Array new:256.
"/      pixel0bytes := ByteArray uninitializedNew:256.
"/      pixel1bytes := ByteArray uninitializedNew:256.
"/
"/      "extract dither patterns and values to use for 1/0 bits
"/       in those from the dithercolors"
"/
"/      1 to:256 do:[:i |
"/          clr := (map at:i) on:aDevice.
"/          ditherPattern := clr ditherForm.
"/
"/          ditherPattern isNil ifTrue:[
"/              patterns at:i put:#[2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111].
"/              pixel0bytes at:i put:clr colorId.
"/              pixel1bytes at:i put:clr colorId
"/          ] ifFalse:[
"/              patterns at:i put:(ditherPattern bits).
"/              pixel0bytes at:i put:(ditherPattern colorMap at:1) colorId.
"/              pixel1bytes at:i put:(ditherPattern colorMap at:2) colorId.
"/          ].
"/      ].
"/
"/      srcIndex := 1.
"/      dstIndex := 1.
"/      mask := 16r80.
"/      outBits := 0.
"/      patternOffset := 1.
"/      1 to:h do:[:dstY |
"/          last := nil.
"/          1 to:w do:[:dstX |
"/              v := bytes at:srcIndex.                             "pixel value"
"/              srcIndex := srcIndex + 1.
"/
"/              v == last ifFalse:[
"/                  index := v + 1.                                "index into map"
"/
"/                  patternBytes := patterns at:index.             "dither pattern for color"
"/                  patternBits := patternBytes at:patternOffset.  "dither row"
"/                  p0 := pixel0bytes at:index.                         "value for 0-dither bit"
"/                  p1 := pixel1bytes at:index.                         "value for 1-dither bit"
"/                  last := v.
"/              ].
"/              outBits := outBits bitShift:1.
"/              (patternBits bitAnd:mask) == 0 ifTrue:[
"/                  outBits := outBits bitOr:p0.
"/              ] ifFalse:[
"/                  outBits := outBits bitOr:p1
"/              ].
"/              mask := mask bitShift:-1.
"/              mask == 0 ifTrue:[
"/                  mask := 16r80.
"/                  formBytes at:dstIndex put:outBits.
"/                  dstIndex := dstIndex + 1.
"/                  outBits := 0
"/              ]
"/          ].
"/          mask == 16r80 ifFalse:[
"/              dstIndex := dstIndex + 1.
"/              mask := 16r80.
"/              outBits := 0
"/          ].
"/          patternOffset := patternOffset + 1.
"/          patternOffset == 9 ifTrue:[
"/              patternOffset := 1
"/          ]
"/      ].
"/      f := Form width:w height:h fromArray:formBytes.
"/      ^ f
    ].

    depth == 2 ifTrue:[
	^ self dither2PlaneUsingMap:map on:aDevice.

"/      formBytes := ByteArray uninitializedNew:(w * 2 + 7 // 8) * h.
"/      patterns := Array new:256.
"/      pixel0bytes := ByteArray uninitializedNew:256.
"/      pixel1bytes := ByteArray uninitializedNew:256.
"/
"/      "extract dither patterns and values to use for 1/o bits
"/       in those from the dithercolors"
"/
"/      1 to:256 do:[:i |
"/          clr := (map at:i) on:aDevice.
"/          ditherPattern := clr ditherForm.
"/
"/          ditherPattern isNil ifTrue:[
"/              patterns at:i put:#[2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111
"/                                  2r11111111].
"/              pixel0bytes at:i put:clr colorId.
"/              pixel1bytes at:i put:clr colorId
"/          ] ifFalse:[
"/              patterns at:i put:(ditherPattern bits).
"/              pixel0bytes at:i put:(ditherPattern colorMap at:1) colorId.
"/              pixel1bytes at:i put:(ditherPattern colorMap at:2) colorId.
"/          ].
"/      ].
"/
"/      srcIndex := 1.
"/      dstIndex := 1.
"/      mask := 16r80.
"/      outBits := 0.
"/      patternOffset := 1.
"/      1 to:h do:[:dstY |
"/          last := nil.
"/          1 to:w do:[:dstX |
"/              v := bytes at:srcIndex.                             "pixel value"
"/              srcIndex := srcIndex + 1.
"/
"/              v == last ifFalse:[
"/                  index := v + 1.                                "index into map"
"/
"/                  patternBytes := patterns at:index.             "dither pattern for color"
"/                  patternBits := patternBytes at:patternOffset.  "dither row"
"/                  p0 := pixel0bytes at:index.                         "value for 0-dither bit"
"/                  p1 := pixel1bytes at:index.                         "value for 1-dither bit"
"/                  last := v.
"/              ].
"/              outBits := outBits bitShift:2.
"/
"/              (patternBits bitAnd:mask) == 0 ifTrue:[
"/                  outBits := outBits bitOr:p0.
"/              ] ifFalse:[
"/                  outBits := outBits bitOr:p1
"/              ].
"/              mask := mask bitShift:-1.
"/              mask == 16r08 ifTrue:[
"/                  formBytes at:dstIndex put:outBits.
"/                  dstIndex := dstIndex + 1.
"/                  outBits := 0
"/              ] ifFalse:[
"/                  mask == 0 ifTrue:[
"/                      mask := 16r80.
"/                      formBytes at:dstIndex put:outBits.
"/                      dstIndex := dstIndex + 1.
"/                      outBits := 0
"/                  ]
"/              ]
"/          ].
"/          ((mask == 16r80) or:[mask == 16r08]) ifFalse:[
"/              formBytes at:dstIndex put:outBits.
"/              dstIndex := dstIndex + 1.
"/              mask := 16r80.
"/              outBits := 0
"/          ].
"/          patternOffset := patternOffset + 1.
"/          patternOffset == 9 ifTrue:[
"/              patternOffset := 1
"/          ]
"/      ].
"/
"/      f := Form width:w height:h depth:2.
"/      f initGC.
"/      f device drawBits:formBytes depth:2
"/                  width:w height:h x:0 y:0
"/                   into:f id x:0 y:0 width:w height:h with:f gcId.
"/      ^ f
    ].

    "draw each pixel using dither color (let others do the dithering)
     although the code is simple, its very slow"

    f := Form width:width height:height depth:depth on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.

    srcIndex := 1.
    1 to:h do:[:dstY |
	run := 0.
	last := nil.
	1 to:w do:[:dstX |
	    |clr v|
	    v := bytes at:srcIndex.
	    srcIndex := srcIndex + 1.
	    clr := map at:(v + 1).

	    clr == last ifTrue:[
		run := run + 1
	    ] ifFalse:[
		(run ~~ 0) ifTrue:[
		    f fillRectangleX:dstX-run-1 y:dstY-1 width:run height:1.
		].
		run := 1.
		f paint:clr.
		last := clr
	    ].
	].
	f fillRectangleX:width-run y:dstY-1 width:run height:1.
    ].
    ^ f
! !

!Depth8Image methodsFor:'image manipulations'!

flipHorizontal
    "inplace horizontal flip"

    |index  "{Class: SmallInteger }"
     h      "{Class: SmallInteger }"
     w      "{Class: SmallInteger }"
     buffer |

    w := width - 1.
    h := height - 1.

    buffer := ByteArray new:width.

    index := 1.
    0 to:h do:[:row |
	buffer replaceFrom:1 to:width with:bytes startingAt:index.
	buffer reverse.
	bytes replaceFrom:index to:index+w with:buffer startingAt:1.
	index := index + w + 1.
    ].
    "flush device info"
    self restored
!

magnifyRowFrom:srcBytes offset:srcStart  
	  into:dstBytes offset:dstStart factor:mX

    "magnify a single pixel row - can only magnify by integer factors.
     Specially tuned for factors 2,3 and 4."

%{
    REGISTER unsigned char *srcP, *dstP;
    REGISTER unsigned char _byte;
    int _mag;
    REGISTER int i;
    int _pixels;
    OBJ w = _INST(width);

    if (__bothSmallInteger(srcStart, dstStart)
     && __bothSmallInteger(w, mX)
     && __isByteArray(srcBytes) && __isByteArray(dstBytes)) {
	_mag = _intVal(mX);
	srcP = _ByteArrayInstPtr(srcBytes)->ba_element - 1 + _intVal(srcStart);
	dstP = _ByteArrayInstPtr(dstBytes)->ba_element - 1 + _intVal(dstStart);
	_pixels = _intVal(w);

	switch (_mag) {
	    case 1:
		break;

	    case 2:
		/* special code for common case */
		while (_pixels--) {
		    _byte = *srcP++;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		}
		break;

	    case 3:
		/* special code for common case */
		while (_pixels--) {
		    _byte = *srcP++;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		}
		break;

	    case 4:
		/* special code for common case */
		while (_pixels--) {
		    _byte = *srcP++;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		    *dstP++ = _byte;
		}
		break;

	    default:
		while (_pixels--) {
		    _byte = *srcP++;
		    for (i=_mag; i>0; i--)
			*dstP++ = _byte;
		}
		break;
	}
	RETURN (self);
    }
%}
.
    self primitiveFailed
!

hardMagnifiedBy:extent
    "return a new image magnified by extent, aPoint.
     This is the general magnification method, handling non-integral values"

    |mX mY
     newWidth  "{ Class: SmallInteger }"
     newHeight "{ Class: SmallInteger }"
     w         "{ Class: SmallInteger }"
     h         "{ Class: SmallInteger }"
     newImage newBytes
     value     "{ Class: SmallInteger }"
     srcRowIdx "{ Class: SmallInteger }"
     srcIndex  "{ Class: SmallInteger }"
     dstIndex  "{ Class: SmallInteger }"|

    mX := extent x.
    mY := extent y.
    ((mX < 0) or:[mY < 0]) ifTrue:[^ nil].
    ((mX = 1) and:[mY = 1]) ifTrue:[^ self].

    newWidth := (width * mX) truncated.
    newHeight := (height * mY) truncated.

    newBytes := ByteArray uninitializedNew:(newWidth * newHeight).

    newImage := self species new.
    newImage bits:newBytes.
    newImage width:newWidth.
    newImage height:newHeight.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:#(8).
    newImage colorMap:colorMap copy.

    "walk over destination image fetching pixels from source image"

    mY := mY asFloat.
    mX := mX asFloat.
%{
    unsigned char *_dstP = _ByteArrayInstPtr(newBytes)->ba_element;
    unsigned char *_srcP = _ByteArrayInstPtr(_INST(bytes))->ba_element;
    unsigned char *_srcRowP;
    int _width = _intVal(_INST(width));
    int _w = _intVal(newWidth) - 1;
    int _h = _intVal(newHeight) - 1;
    int _row, _col;
    double _mX = _floatVal(mX);
    double _mY = _floatVal(mY);

    for (_row = 0; _row <= _h; _row++) {
	_srcRowP = _srcP + (_width * (int)((double)_row / _mY));
	for (_col = 0; _col <= _w; _col++) {
	    *_dstP++ = _srcRowP[(int)((double)_col / _mX)];
	}
    }
%}
.
"/   the above C-code is equivalent to:
"/
"/    dstIndex := 1.
"/    w := newWidth - 1.
"/    h := newHeight - 1.
"/    0 to:h do:[:row |
"/        srcRowIdx := (width * (row // mY)) + 1.
"/        0 to:w do:[:col |
"/            srcIndex := srcRowIdx + (col // mX).
"/            value := bytes at:srcIndex.
"/            newBytes at:dstIndex put:value.
"/            dstIndex := dstIndex + 1
"/        ]
"/    ].
"/

    ^ newImage
! !
